//===--- NetSession.h - Network Session Manager -----------------*- C++ -*-===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// Handles all networking tasks required for network distributed compiling.
// Support message based communication via blocked mode BSD socket.
//
//===----------------------------------------------------------------------===//

#ifndef NET_SESSION_H_
#define NET_SESSION_H_
#include <sys/socket.h>
#include <netinet/in.h>

#include "llvm/Support/MemoryBuffer.h"
#include <string>
#include <vector>
#include <sstream>

namespace clang {

enum RemoteFileType {
  RemoteFileNotExistsT,
  RemoteFileT,
  RemoteStdOutT,
  RemoteStdErrT
};

enum RemoteActions {
  RemoteError,
  RemoteComplete,
  RemoteProcessSerializedFile,
  RemoteProcessInputFile,
  RemoteDiagnostic,
  RemoteFileRequest,
  RemoteFile
};

class NetSession {
private:
  std::string ErrorString;

  long HostAddress;
  int ClientSocket;

  struct hostent* HostInfo;
  struct sockaddr_in Address;
  int AddressSize;
  int ServerSocket;
  
  bool ClientMode;

  struct PacketFile {
    std::string FileName;
    RemoteFileType FileType;
    std::ostringstream *FileStream;
    int FileSize;
  };
  
  llvm::Serializer *PacketHeaderSer;
  std::vector<unsigned char> *PacketHeaderBuf;
  std::vector<PacketFile> PacketFiles;
  
public:
  int HostPort;
  int QueueSize;
  
public:
  NetSession();
  NetSession(int pClientSocket);
  
  bool StartServer();
  int AcceptConnection();
  
  bool ConnectToServer();
  
  bool Close();
  
  std::string& GetErrorString();

  void PacketAddFile(const char *FName, RemoteFileType FType = RemoteFileT);
  void PacketAddFile(const char *FName,
                     std::ostringstream *pFileStream,
                     RemoteFileType FType = RemoteFileT);
  void SetPacketHeader(llvm::Serializer *pPacketHeaderSer,
                       std::vector<unsigned char> *pPacketHeaderBuf);
  void ClearPacket();
  bool SendPacket();
  llvm::MemoryBuffer* GetPacket();
};

} // end namspace clang

#endif
