//===--- NetSession.cpp - Network Session Manager ---------------*- C++ -*-===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
//  This file defines the NetSession object.
//
//===----------------------------------------------------------------------===//

//#include <sys/types.h>
//#include <string.h>
//#include <unistd.h>
//#include <stdio.h>

#include <netdb.h>
#include <fstream>
#include "llvm/System/Path.h"
#include "llvm/ADT/OwningPtr.h"
#include "llvm/Bitcode/ReaderWriter.h"
#include "llvm/Bitcode/Serialize.h"

#include "NetSession.h"

#define SOCKET_ERROR     -1

using namespace clang;

NetSession::NetSession() {
  AddressSize = sizeof(struct sockaddr_in);
  HostPort = 50000;
  ClientMode = false;
  ClearPacket();
}

NetSession::NetSession(int pClientSocket) {
  AddressSize = sizeof(struct sockaddr_in);
  HostPort = 50000;
  ClientSocket = pClientSocket;
  ClientMode = true;
  ClearPacket();
}

bool NetSession::StartServer() {

  ServerSocket = socket(AF_INET,SOCK_STREAM,0);

  if (ServerSocket == SOCKET_ERROR) {
    ErrorString = "NetSession::StartServer: Could not make a socket";
    return false;
  }

  Address.sin_addr.s_addr = INADDR_ANY;
  Address.sin_port = htons(HostPort);
  Address.sin_family = AF_INET;

  if (bind(ServerSocket, (struct sockaddr*)&Address, sizeof(Address))
      == SOCKET_ERROR) {
    ErrorString = "NetSession::StartServer: Could not connect to host";
    return false;
  }
  
  getsockname(ServerSocket, (struct sockaddr*)&Address,
             (socklen_t*)&AddressSize);

  if (listen(ServerSocket, QueueSize) == SOCKET_ERROR) {
    ErrorString = "NetSession::StartServer: Could not listen";
    return false;
  }

  return true;
}

int NetSession::AcceptConnection() {

  int Socket = accept(ServerSocket, (struct sockaddr*)&Address,
                      (socklen_t *)&AddressSize);

  // TODO: add protocol check and version check
  return Socket;
}

bool NetSession::ConnectToServer() {

  std::string HostName = "hcube";
  // TODO
  // read the configuration file
  // store server list
  // randomize server list

  // iterate over server list and connect to first available one
  // TODO
  
  // make a socket
  ClientSocket = socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);

  if (ClientSocket == SOCKET_ERROR) {
    ErrorString = "NetSession::ConnectToServer: Could not make a socket";
    return false;
  }

  // get IP address from name
  HostInfo = gethostbyname(HostName.c_str());
  // copy address into long
  memcpy(&HostAddress,HostInfo->h_addr,HostInfo->h_length);

  // fill address struct
  Address.sin_addr.s_addr = HostAddress;
  Address.sin_port = htons(HostPort);
  Address.sin_family = AF_INET;

  // connect to host
  if (connect(ClientSocket, (struct sockaddr*)&Address, sizeof(Address))
      == SOCKET_ERROR) {
    ErrorString = "NetSession::ConnectToServer: Could not connect";
    return false;
  }
  return true;
}

bool NetSession::Close() {
  int Res;

  if (ClientMode)
    Res = close(ClientSocket);
  else
    Res = close(ServerSocket);
	
  if (Res == SOCKET_ERROR) {
    ErrorString = "NetSession::Close: Could not close socket";
    return false;
  }
  return true;
}

std::string& NetSession::GetErrorString() {
  return ErrorString;
}

void NetSession::PacketAddFile(const char *FName,
                               std::ostringstream *pFileStream,
                               RemoteFileType FType) {
  PacketFile PF;
  PF.FileName = FName;
  PF.FileType = FType;
  if (FType == RemoteFileNotExistsT) {
    PF.FileStream = 0;
    PF.FileSize = 0;
  }
  else {
    PF.FileStream = pFileStream;
    pFileStream->rdbuf()->pubseekoff(0,std::ios_base::beg,std::ios_base::in);
    pFileStream->rdbuf()->pubseekoff(0,std::ios_base::end,std::ios_base::out);
    PF.FileSize = pFileStream->rdbuf()->in_avail();
  }

  PacketFiles.push_back(PF);
}

void NetSession::PacketAddFile(const char *FName, RemoteFileType FType) {
  PacketFile PF;
  PF.FileName = FName;
  PF.FileType = FType;
  PF.FileStream = NULL;
  if (FType == RemoteFileNotExistsT) {
    PF.FileSize = 0;
  }
  else {
    llvm::sys::PathWithStatus File(FName);
    PF.FileSize = File.getFileStatus()->getSize();
  }

  PacketFiles.push_back(PF);
}

void NetSession::SetPacketHeader(llvm::Serializer *pPacketHeaderSer,
                                 std::vector<unsigned char> *pPacketHeaderBuf) {
  PacketHeaderSer = pPacketHeaderSer;
  PacketHeaderBuf = pPacketHeaderBuf;
}

void NetSession::ClearPacket() {
  PacketHeaderSer = NULL;
  PacketHeaderBuf = NULL;
  PacketFiles.clear();
}

bool NetSession::SendPacket() {
  
  // Make Bitstream section
  llvm::OwningPtr<std::vector<unsigned char> > TmpHeaderBuf;
  llvm::OwningPtr<llvm::BitstreamWriter> TmpHeaderStream;
  llvm::OwningPtr<llvm::Serializer> TmpHeaderSer;
  
  std::vector<unsigned char> *HeaderBuf = PacketHeaderBuf;
  llvm::Serializer *HeaderSer = PacketHeaderSer;
  u_long NSize, FilesSize = 0;
  
  // Make Bitstream buffer if necessary.
  if (PacketHeaderBuf == NULL) {
    TmpHeaderBuf.reset(new std::vector<unsigned char>);
    TmpHeaderStream.reset(new llvm::BitstreamWriter(*TmpHeaderBuf));
    TmpHeaderSer.reset(new llvm::Serializer (*TmpHeaderStream));  
    HeaderBuf = TmpHeaderBuf.get();
    HeaderSer = TmpHeaderSer.get();
    
    HeaderSer->EnterBlock(0);  
  }

  for (std::vector<PacketFile>::iterator i = PacketFiles.begin() ;
       i != PacketFiles.end() ; i++) {
    // Make file header: RemoteAction, FileName, FileSize,
    //                   RemoteFileType, FileOffset
    HeaderSer->EmitInt(RemoteFile);
    HeaderSer->EmitCStr((*i).FileName.c_str());
    HeaderSer->EmitInt((*i).FileSize);
    HeaderSer->EmitInt((*i).FileType);
    HeaderSer->EmitInt(FilesSize);
    FilesSize += (*i).FileSize;

    // DEBUG
    fprintf(stderr, " * Add file: '%s' size: %d type: %d\n",
            (*i).FileName.c_str(),
            (*i).FileSize,
            (*i).FileType);
  }
  // Close Bitstream section.
  HeaderSer->ExitBlock();
  HeaderSer->FlushRecord();
  
  // DEBUG
  fprintf(stderr, " * sent packet size: %ld\n",
          sizeof(u_long) + HeaderBuf->size() + FilesSize);

  const unsigned int BufferSize = 60*1024;
  unsigned char Buffer[BufferSize];
  unsigned int BufferLen = 3 + 2*sizeof(u_long);
  
  Buffer[0] = 'C';
  Buffer[1] = 'F';
  Buffer[2] = 'E';
  // Packet size.
  NSize = htonl(sizeof(u_long) + HeaderBuf->size() + FilesSize);
  memcpy(&Buffer[3], &NSize, sizeof(u_long));

  // Bitstream section size.
  NSize = htonl(HeaderBuf->size());
  memcpy(&Buffer[3+sizeof(u_long)], &NSize, sizeof(u_long));

  // Send bitcode section.
  for (std::vector<unsigned char>::iterator i = HeaderBuf->begin() ;
       i != HeaderBuf->end() ; ++i) {
    Buffer[BufferLen++] = (*i);

    if (BufferLen >= BufferSize) {
      write(ClientSocket, Buffer, BufferLen);
      BufferLen = 0;
    }
  }
  // Send files data.
  for (std::vector<PacketFile>::iterator i = PacketFiles.begin() ;
       i != PacketFiles.end() ; i++) {

    if ((*i).FileType == RemoteFileNotExistsT)
	  continue;

    // Send file data.
    std::streambuf *PBuf;
    std::ifstream *FileStream;

    if ((*i).FileStream == NULL) {
      FileStream = new std::ifstream((*i).FileName.c_str());
      PBuf = FileStream->rdbuf();
    }
    else
      PBuf = (*i).FileStream->rdbuf();

    for (int j = (*i).FileSize ; j > 0 ; --j) {
      Buffer[BufferLen++] = PBuf->sbumpc();

      if (BufferLen >= BufferSize) {
        write(ClientSocket, Buffer, BufferLen);
        BufferLen = 0;
      }
    }
    if ((*i).FileStream == NULL) {
      FileStream->close();
      delete FileStream;
    }
  }

  if (BufferLen > 0)
    write(ClientSocket, Buffer, BufferLen);

  return true;
}

llvm::MemoryBuffer* NetSession::GetPacket() {
  char ID[3]; // CFE
  u_long Size, NSize;

  read(ClientSocket, &ID, 3);
  if (ID[0]!='C' || ID[1]!='F' || ID[2]!='E' )
    return NULL;
  read(ClientSocket, &NSize, sizeof(u_long));

  Size = ntohl(NSize);

  // DEBUG
  fprintf(stderr, " * incoming packet size: %ld\n", Size);

  llvm::MemoryBuffer *Data = llvm::MemoryBuffer::getNewUninitMemBuffer(Size);
  read(ClientSocket, (void*)Data->getBufferStart(), Size);

  // DEBUG
  std::ofstream outfile ("packetGot.bin",std::ofstream::binary);
  outfile.write(ID,3);
  outfile.write((char*)&NSize,sizeof(u_long));
  outfile.write(Data->getBufferStart(), Size);
  outfile.close();
  
  return Data;
}
