//===--- clangserver.cpp - C-Language Front-end compile server-------------===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
//  This utility may be invoked in the following manner:
//   clangserver --help                - Output help info.
//
//===----------------------------------------------------------------------===//

#include "clang.h"
#include "ASTConsumers.h"
#include "DiagnosticBuffer.h"
#include "HTMLDiagnostics.h"
#include "NetSession.h"
#include "clang/Analysis/PathDiagnostic.h"
#include "clang/AST/TranslationUnit.h"
#include "clang/CodeGen/ModuleBuilder.h"
#include "clang/Sema/ParseAST.h"
#include "clang/AST/ASTConsumer.h"
#include "clang/Parse/Parser.h"
#include "clang/Lex/HeaderSearch.h"
#include "clang/Basic/FileManager.h"
#include "clang/Basic/SourceManager.h"
#include "clang/Basic/TargetInfo.h"
#include "llvm/Module.h"
#include "llvm/ADT/SmallPtrSet.h"
#include "llvm/Bitcode/Serialize.h"
#include "llvm/Bitcode/Deserialize.h"
#include "llvm/Bitcode/ReaderWriter.h"
#include "llvm/Support/CommandLine.h"
#include "llvm/Support/MemoryBuffer.h"
#include "llvm/System/Signals.h"
#include "llvm/Config/config.h"
#include "llvm/ADT/OwningPtr.h"
#include "llvm/System/Path.h"
#include <memory>
#include <fstream>
#include <sstream>
#include <algorithm>
using namespace clang;

//===----------------------------------------------------------------------===//
// Global options.
//===----------------------------------------------------------------------===//

static llvm::cl::opt<bool>
Verbose("v", llvm::cl::desc("Enable verbose output"));
static llvm::cl::opt<bool>
Stats("print-stats", 
      llvm::cl::desc("Print performance metrics and statistics"));

//===----------------------------------------------------------------------===//
// Server options.
//===----------------------------------------------------------------------===//

static llvm::cl::opt<unsigned>
QueueSize("queuesize",
      llvm::cl::desc("Server request queue size"));
	  
//===----------------------------------------------------------------------===//
// Main driver
//===----------------------------------------------------------------------===//

/// CreateASTConsumer - Create the ASTConsumer for the corresponding program
///  action.  These consumers can operate on both ASTs that are freshly
///  parsed from source files as well as those deserialized from Bitcode.
static ASTConsumer* CreateASTConsumer(const std::string& InFile,
                                      Diagnostic& Diag, FileManager& FileMgr, 
                                      const LangOptions& LangOpts,
                                      Preprocessor *PP,
                                      llvm::Module *&DestModule,
                                      ProgActions ProgAction,
                                      std::ostream *StdOut) {

  switch (ProgAction) {
  default:
    fprintf(stderr, "Unexpected program action!\n");
    break;
      
  case ParseSyntaxOnly:              // -fsyntax-only, no output
    return new ASTConsumer();
      
  case ASTPrint:
    return CreateASTPrinter(StdOut);
      
  case ASTDump:
    return CreateASTDumper(StdOut);
      
  case EmitHTML:
    //return CreateHTMLPrinter(StdOut, Diag, PP, PPF);
    break;

  case EmitLLVM:
  case EmitBC:
    DestModule = new llvm::Module(InFile);
    //return CreateLLVMCodeGen(Diag, LangOpts, DestModule, GenerateDebugInfo);
    break;

  case SerializeAST:
    // FIXME: Allow user to tailor where the file is written.
    //return CreateASTSerializer(InFile, StdOut, Diag);
    break;
      
  case RewriteObjC:
    //return CreateCodeRewriterTest(InFile, StdOut, Diag, LangOpts);
    break;
      
  case RunAnalysis:
    // Emits more than one file.
    //assert (!AnalysisList.empty());
    //Consumer = CreateAnalysisConsumer(&AnalysisList[0],
    //                                  &AnalysisList[0]+AnalysisList.size(),
    //                                  Diag, PP, PPF, LangOpts,
    //                                  AnalyzeSpecificFunction,
    //                                  OutputFile, VisualizeEG, TrimGraph,
    //                                  AnalyzeAll);
    break;
  }
  return NULL;
}

/// ProcessInputFile - Process a single input file with the specified state.
///
static void ProcessInputFile(Preprocessor &PP, const std::string &InFile,
                             ProgActions ProgAction, std::ostream *StdOut) {
  ASTConsumer* Consumer = NULL;
  llvm::Module *CodeGenModule = 0;
  
  Consumer = CreateASTConsumer(InFile, PP.getDiagnostics(),
                               PP.getFileManager(), PP.getLangOptions(),
                               &PP, CodeGenModule,
							   ProgAction, StdOut);
  
  // This deletes Consumer.
  ParseAST(PP, Consumer, Stats);

  // If running the code generator, finish up now.
  if (CodeGenModule) {
    if (ProgAction == EmitLLVM) {
      CodeGenModule->print(StdOut);
    } else {
      assert(ProgAction == EmitBC);
      llvm::WriteBitcodeToFile(CodeGenModule, *StdOut);
    }
    delete CodeGenModule;
  }
  /*
  if (Stats) {
    fprintf(stderr, "\nSTATISTICS FOR '%s':\n", InFile.c_str());
    PP.PrintStats();
    PP.getIdentifierTable().PrintStats();
    PP.getHeaderSearchInfo().PrintStats();
    PP.getSourceManager().PrintStats();
    fprintf(stderr, "\n");
  }
  */
}

static void ProcessSerializedFile(llvm::MemoryBuffer& BufferIn,
                                  Diagnostic& Diag,
                                  FileManager& FileMgr,
                                  ProgActions ProgAction,
                                  std::ostream *StdOut) {

  llvm::OwningPtr<TranslationUnit> TU(ReadASTBitcodeBuffer(BufferIn, FileMgr));
  std::string InFile = BufferIn.getBufferIdentifier();

  if (!TU) {
    //ErrorMsg = "error: file '"+InFile+"' could not be deserialized";
    return;
  }
  
  // Observe that we use the source file name stored in the deserialized
  // translation unit, rather than InFile.
  llvm::Module *CodeGenModule;

  ASTConsumer *Consumer = CreateASTConsumer(InFile, Diag, FileMgr,
                                            TU->getLangOptions(), 0,
                                            CodeGenModule, ProgAction, StdOut);
  
  if (!Consumer) {      
    //ErrorMsg = "Unsupported program action with serialized ASTs!";
    return;
  }
  
  Consumer->Initialize(TU->getContext());
  
  // FIXME: We need to inform Consumer about completed TagDecls as well.
  for (TranslationUnit::iterator I=TU->begin(), E=TU->end(); I!=E; ++I)
    Consumer->HandleTopLevelDecl(*I);

  delete Consumer;

  if (CodeGenModule) {
    if (ProgAction == EmitLLVM) {
      CodeGenModule->print(StdOut);
    } else {
      assert(ProgAction == EmitBC);
      llvm::WriteBitcodeToFile(CodeGenModule, *StdOut);
    }
    delete CodeGenModule;
  }
}

static llvm::MemoryBuffer* GetRemoteFileBuffer(llvm::Deserializer &PacketIn,
                                               unsigned char *FilesData) {
  // File header: RemoteAction, FileName, FileSize,
  //              RemoteFileType, FileOffset
  RemoteActions RemoteActions = (clang::RemoteActions)PacketIn.ReadInt();
  std::string FName = PacketIn.ReadCStr();
  int FSize = PacketIn.ReadInt();
  RemoteFileType FType = (clang::RemoteFileType)PacketIn.ReadInt();
  int FOffset = PacketIn.ReadInt();

  return llvm::MemoryBuffer::getMemBuffer((const char*)FilesData+FOffset,
                                          (const char*)FilesData+FOffset+FSize,
                                          FName.c_str());
}

static void ServeRequest(int Connection) {

  NetSession Session(Connection);
  
  llvm::MemoryBuffer *FileBuffer;
  std::ostringstream StdOut(std::ios_base::out|
                            std::ios_base::in|
                            std::ios_base::binary);
  std::string InFile;
  //std::string ErrorMsg;
  
  // Read incoming request packet.
  llvm::MemoryBuffer *BufferIn = Session.GetPacket();
  if (BufferIn == NULL) {
    fprintf(stderr, "Unrecognized incoming packet\n");
    Session.Close();
    return;
  }
  
  unsigned char *BufPtr = (unsigned char *) BufferIn->getBufferStart();
  u_long HeaderSize = ntohl(*((u_long*)BufPtr));
  BufPtr += sizeof(u_long);
  unsigned char *FilesData = BufPtr + HeaderSize;
  llvm::BitstreamReader StreamIn(BufPtr, BufPtr+HeaderSize);

  llvm::Deserializer PacketIn(StreamIn);
  PacketIn.SkipToBlock(0);
  
  // Parse incoming request packet.
  RemoteActions RemoteAction = (clang::RemoteActions)PacketIn.ReadInt();
  ProgActions ProgAction = (clang::ProgActions)PacketIn.ReadInt();
  std::string OutputFile = PacketIn.ReadCStr();

  // Setup clang

  // Configure our handling of diagnostics.
  llvm::OwningPtr<DiagnosticClient> DiagClient(new DiagnosticBuffer());
  Diagnostic Diags(*DiagClient);

  // Create a file manager object to provide access to and cache the filesystem.
  FileManager FileMgr;

  if (RemoteAction == RemoteProcessSerializedFile) {
    // Get file buffer from packet.
    FileBuffer = GetRemoteFileBuffer(PacketIn,FilesData);
    InFile = FileBuffer->getBufferIdentifier();
    ProcessSerializedFile(*FileBuffer, Diags, FileMgr, ProgAction, &StdOut);
    delete FileBuffer;
  }
  else {
    // Get information about the target being compiled for.
    std::string Triple = PacketIn.ReadCStr();
    llvm::OwningPtr<TargetInfo> Target(TargetInfo::CreateTargetInfo(Triple));
  
    // Initialize language options from packet.
    LangOptions LangInfo;
    PacketIn.Read(LangInfo);

    // Get predefines from packet.
    std::string Predefines = PacketIn.ReadCStr();

    // Get MainFile from packet.
    FileBuffer = GetRemoteFileBuffer(PacketIn,FilesData);
    InFile = FileBuffer->getBufferIdentifier();

    SourceManager SourceMgr;
    HeaderSearch HeaderInfo(FileMgr);
    Preprocessor PP(Diags, LangInfo, *Target, SourceMgr, HeaderInfo);
    PP.setPredefines(Predefines);
    // SourceManager takes ownership of FileBuffer
    SourceMgr.createMainFileIDForMemBuffer(FileBuffer);

    ProcessInputFile(PP, InFile, ProgAction, &StdOut);

    if (Stats)
      SourceMgr.PrintStats();

  } 

  // Write out the output.
  StdOut.rdbuf()->pubseekoff(0,std::ios_base::beg,std::ios_base::in);
  StdOut.rdbuf()->pubseekoff(0,std::ios_base::end,std::ios_base::out);

  // DEBUG
  fprintf(stderr, " * ProgAction: %d\n", ProgAction);
  fprintf(stderr, " * OutputFile: %s\n", OutputFile.c_str());
  fprintf(stderr, " * OutputSize: %d\n", StdOut.rdbuf()->in_avail());
  fprintf(stderr, " * StdOut: %s\n", StdOut.str().c_str());

  if (StdOut.rdbuf()->in_avail() > 0) {
    // Create StdOut output or OutputFile
    
	// Some dirty hack to modify output file name.
    if ((ProgAction == EmitLLVM || ProgAction == EmitBC) && OutputFile == "") {
      llvm::sys::Path Path(InFile);
      Path.eraseSuffix();
      if (ProgAction == EmitLLVM)
        Path.appendSuffix("ll");
      else if (ProgAction == EmitBC)
        Path.appendSuffix("bc");
      OutputFile = Path.toString();
    }
	
    if (OutputFile == "" || OutputFile == "-")
      Session.PacketAddFile(OutputFile.c_str(), &StdOut, RemoteStdOutT);
    else
      Session.PacketAddFile(OutputFile.c_str(), &StdOut, RemoteFileT);
  }
  
  std::vector<unsigned char> BufferOut;
  llvm::BitstreamWriter StreamOut(BufferOut);
  llvm::Serializer PacketOut(StreamOut);  
  PacketOut.EnterBlock(0);  

  // TODO  
  // write diagnostic results to packet

  // Close result packet with RemoteComplete
  PacketOut.EmitInt(RemoteComplete);  
  
  // send result packet
  Session.SetPacketHeader(&PacketOut, &BufferOut);
  Session.SendPacket();
  /*
  if (Stats) {
    FileMgr.PrintStats();
    fprintf(stderr, "\n");
  }
  */
  delete BufferIn;
}

int main(int argc, char **argv) {
  llvm::cl::ParseCommandLineOptions(argc, argv, " llvm clang cfe server\n");
  llvm::sys::PrintStackTraceOnErrorSignal();
  
  NetSession Session;
  Session.StartServer();

  while (true) {
    int Connection = Session.AcceptConnection();
    // TODO
    // start compiler task (create a thread later)
    ServeRequest(Connection);
  }

  Session.Close();
}
